#!/usr/bin/python
"""Generates unique, default names for a volume group and logical volume"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils import facts

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: lvm_gensym
short_description: Generate default names for lvm variables
version_added: "2.4"
description: 
    - "Module accepts two input strings consisting of a file system type and 
       a mount point path, and outputs names based on system information"
options:
    fs_type:
        description:
            - String describing the desired file system type 
        required: true 
    mount:
        description:
            - String describing the mount point path 
        required: true
author: 
    - Tim Flannagan (tflannag@redhat.com)
'''

EXAMPLES = '''
- name: Generate names 
  lvm_gensym:
    fs_type: "{{ fs_type }}"
    mount: "{{ mount_point }}"
  register: lvm_results
  when: lvm_vg == "" and mount_point != "" and fs_type != ""
'''

RETURN = '''
vg_name:
    description: The default generated name for an unspecified volume group
    type: str 
'''

RETURN = '''
vg_name:
    description: The default generated name for an unspecified volume group
    type: str 

lv_name:
    description: The default generated name for an unspecified logical volume 
    type: str
'''


def get_os_name():
    """Search the host file and return the name in the ID column"""
    for line in open('/etc/os-release').readlines():
        if not line.find('ID='):
            os_name = line[3:]
            break

    os_name = os_name.replace('\n', '').replace('"', '')
    return os_name

def name_is_unique(name, used_names):
    """Check if name is contained in the used_names list and return boolean value"""
    if name not in used_names:
        return True

    return False

def get_unique_name_from_base(base_name, used_names):
    """Generate a unique name given a base name and a list of used names, and return that unique name"""
    counter = 0
    while not name_is_unique(base_name, used_names):
        if counter == 0:
            base_name = base_name + '_' + str(counter)
        else:
            base_name = base_name[:-2] + '_' + str(counter)
        counter += 1

    return base_name


def get_vg_name_base(host_name, os_name):
    """Return a base name for a volume group based on the host and os names"""
    if host_name != None and len(host_name) != 0:
        vg_default = os_name + '_' + host_name
    else:
        vg_default = os_name

    return vg_default


def get_vg_name(host_name, lvm_facts):
    """Generate a base volume group name, verify its uniqueness, and return that unique name"""
    used_vg_names = lvm_facts['vgs'].keys()
    os_name = get_os_name()
    name = get_vg_name_base(host_name, os_name)

    return get_unique_name_from_base(name, used_vg_names)

def get_lv_name_base(fs_type, mount_point):
    """Return a logical volume base name using given parameters"""
    if 'swap' in fs_type.lower():
        lv_default = 'swap'
    elif mount_point.startswith('/'):
        if mount_point == '/':
            lv_default = 'root'
        else:
            lv_default = mount_point[1:].replace('/', '_')
    else:
        lv_default = 'lv'

    return lv_default


def get_lv_name(fs_type, mount_point, lvm_facts):
    """Return a unique logical volume name based on specified file system type, mount point, and system facts"""
    used_lv_names = lvm_facts['lvs'].keys()
    name = get_lv_name_base(fs_type, mount_point)

    return get_unique_name_from_base(name, used_lv_names)

def run_module():
    """Setup and initialize all relevant ansible module data"""
    module_args = dict(
        mount=dict(type='str', required=True),
        fs_type=dict(type='str', required=True)
    )

    result = dict(
        changed=False,
        vg_name='',
        lv_name=''
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    lvm_facts = facts.ansible_facts(module)['lvm']
    host_name = facts.ansible_facts(module)['nodename'].lower().replace('.', '_').replace('-', '_')

    result['lv_name'] = get_lv_name(module.params['fs_type'], module.params['mount'], lvm_facts)
    result['vg_name'] = get_vg_name(host_name, lvm_facts)

    if result['lv_name'] != '' and result['vg_name'] != '':
        module.exit_json(**result)
    else:
        module.fail_json(msg="Unable to initialize both group and volume names")

def main():
    run_module()

if __name__ == '__main__':
    main()



#!/usr/bin/python

DOCUMENTATION = '''
---
module: find_unused_disk
short_description: Gets unused disks
description:
    - Disks are considered in ascending alphanumeric sorted order.
    - Disks that meet all conditions are considered 'empty' and returned (using kernel device name) in a list.
        - 1. No known signatures exist on the disk, with the exception of partition tables.
        - 2. If there is a partition table on the disk, it contains no partitions.
        - 3. The disk has no holders to eliminate the possibility of it being a multipath or dmraid member device.
        - 4. Device can be opened with exclusive access to make sure no other software is using it.
    - If no disks meet all criteria, "Unable to find unused disk" will be returned.
    - Number of returned disks defaults to first 10, but can be specified with 'max_return' argument.
author: Eda Zhou (@edamamez)
options:
    option-name: max_return
    description: Sets the maximum number of unused disks to return.
    default: 10
    type: int

    option-name: min_size
    description: Specifies the minimum disk size to return an unused disk.
    default: 0
    type: str
'''

EXAMPLES = '''
- name: test finding first unused device module
  hosts: localhost
  tasks:
    - name: run module
      find_unused_disk:
        min_size: '10g'
      register: testout

    - name: dump test output
      debug:
        msg: '{{ testout }}'
'''

RETURN = '''
disk_name:
    description: Information about unused disks
    returned: On success
    type: complex
    contains:
        disks:
            description: Unused disk(s) that have been found
            returned: On success
            type: list
            samples: ["sda1", "dm-0", "dm-3"]
                     ["sda"]
        none:
            description: No unused disks were found
            returned: On success
            type: string
            sample: "Unable to find unused disk"
'''


import os
import re

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.size import Size


SYS_CLASS_BLOCK = "/sys/class/block/"

def no_signature(run_command, disk_path):
    """Return true if no known signatures exist on the disk."""
    signatures = run_command(['blkid', '-p', disk_path])
    return not 'UUID' in signatures[1]


def no_holders(disk_path):
    """Return true if the disk has no holders."""
    holders = os.listdir(SYS_CLASS_BLOCK + get_sys_name(disk_path) + '/holders/')
    return len(holders) == 0


def can_open(disk_path):
    """Return true if the device can be opened with exclusive access."""
    try:
        os.open(disk_path, os.O_EXCL)
        return True
    except OSError:
        return False


def get_sys_name(disk_path):
    if not os.path.islink(disk_path):
        return os.path.basename(disk_path)

    node_dir = '/'.join(disk_path.split('/')[-1])
    return os.path.normpath(node_dir + '/' + os.readlink(disk_path))


def get_partitions(disk_path):
    sys_name = get_sys_name(disk_path)
    partitions = list()
    for filename in os.listdir(SYS_CLASS_BLOCK + sys_name):
        if re.match(sys_name + 'p?\d+$', filename):
            partitions.append(filename)

    return partitions


def get_disks(run_command):
    buf = run_command(["lsblk", "-p", "--pairs", "--bytes", "-o", "NAME,TYPE,SIZE,FSTYPE"])[1]
    disks = dict()
    for line in buf.splitlines():
        if not line:
            continue

        m = re.search(r'NAME="(?P<path>[^"]*)" TYPE="(?P<type>[^"]*)" SIZE="(?P<size>\d+)" FSTYPE="(?P<fstype>[^"]*)"', line)
        if m is None:
            print(line)
            continue

        if m.group('type') != "disk":
            continue

        disks[m.group('path')] = {"type": m.group('type'), "size": m.group('size'), "fstype": m.group('fstype')}

    return disks


def run_module():
    """Create the module"""
    module_args = dict(
        max_return=dict(type='int', required=False, default=10),
        min_size=dict(type='str', required=False, default=0)
    )

    result = dict(
        changed=False,
        disks=[]
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    run_command = module.run_command

    for path, attrs in get_disks(run_command).items():
        if attrs["fstype"]:
            continue

        if Size(attrs["size"]).bytes < Size(module.params['min_size']).bytes:
            continue

        if get_partitions(path):
            continue

        if not no_holders(get_sys_name(path)):
            continue

        if not can_open(path):
            continue

        result['disks'].append(os.path.basename(path))
        if len(result['disks']) >= module.params['max_return']:
            break

    if not result['disks']:
        result['disks'] = "Unable to find unused disk"
    else:
        result['disks'].sort()

    module.exit_json(**result)


def main():
    """Execute the module"""
    run_module()


if __name__ == '__main__':
    main()



#!/bin/python2

import re

DECIMAL_FACTOR = 10 ** 3
BINARY_FACTOR = 2 ** 10

# index of the item in the list determines the exponent for size computation
# e.g. size_in_bytes = value * (DECIMAL_FACTOR ** (index(mega)+1)) = value * (1000 ** (1+1))
PREFIXES_DECIMAL = [["k",    "M",    "G",    "T",    "P",    "E",    "Z",     "Y"],
                    ["kilo", "mega", "giga", "tera", "peta", "exa",  "zetta", "yotta"]]
PREFIXES_BINARY  = [["Ki",   "Mi",   "Gi",   "Ti",   "Pi",   "Ei",   "Zi",    "Yi"],
                    ["kibi", "mebi", "gibi", "tebi", "pebi", "exbi", "zebi",  "yobi"]]
SUFFIXES = ["bytes", "byte", "B"]

class Size(object):
    ''' Class for basic manipulation of the sizes in *bytes
    '''

    def __init__(self, value):
        raw_number, raw_units = self._parse_input(str(value))
        self.factor, self.exponent = self._parse_units(raw_units)
        self.number = self._parse_number(raw_number)

        self.units = raw_units

    def _parse_input(self, value):
        ''' splits input string into number and unit parts
            returns number part, unit part
        '''
        m = re.search("^(.*?)([^0-9]*)$", value)

        raw_number = m.group(1).strip()
        if raw_number == "":
            raise ValueError("The string '%s' does not contain size" % value)

        raw_units = m.group(2).strip()

        return raw_number, raw_units

    def _parse_units(self, raw_units):
        '''
            gets string containing size units and
            returns *_FACTOR (BINARY or DECIMAL) and the prefix position (not index!)
            in the PREFIXES_* list
            If no unit is specified defaults to BINARY and Bytes
        '''

        prefix = raw_units
        no_suffix_flag = True
        valid_suffix = False

        # get rid of possible units suffix ('bytes', 'b' or 'B')
        for suffix in SUFFIXES:
            if raw_units.lower().endswith(suffix.lower()):
                no_suffix_flag = False
                prefix = raw_units[:-len(suffix)]
                break

        if prefix == "":
            # no unit was specified, use default
            return BINARY_FACTOR, 0

        # check the list for units
        idx = -1

        for lst in PREFIXES_DECIMAL:
            lower_lst = [x.lower() for x in lst]
            if prefix.lower() in lower_lst:
                valid_suffix = True
                idx = lower_lst.index(prefix.lower())
                used_factor = DECIMAL_FACTOR
                break

        if idx < 0 or no_suffix_flag:
            if no_suffix_flag:
                used_factor = BINARY_FACTOR

            for lst in PREFIXES_BINARY:
                lower_lst = [x.lower() for x in lst]
                if prefix.lower() in lower_lst:
                    valid_suffix = True
                    idx = lower_lst.index(prefix.lower())
                    used_factor = BINARY_FACTOR
                    break

        if idx < 0 or not valid_suffix:
            raise ValueError("Unable to identify unit '%s'" % raw_units)

        return used_factor, idx+1

    def _parse_number(self, raw_number):
        ''' parse input string containing number
            return float
        '''
        return float(raw_number)

    def _get_unit(self, factor, exponent, unit_type=0):
        ''' based on decimal or binary factor and exponent
            obtain and return correct unit
        '''

        if unit_type == 0:
            suffix = "B"
        else:
            suffix = "bytes"

        if exponent == 0:
            return suffix

        if factor == DECIMAL_FACTOR:
            prefix_lst = PREFIXES_DECIMAL[unit_type]
        else:
            prefix_lst = PREFIXES_BINARY[unit_type]
        return prefix_lst[exponent-1] + suffix

    @property
    def bytes(self):
        ''' returns size value in bytes as int
        '''
        return int((self.factor ** self.exponent) * self.number)

    def _format(self, format_str, factor, exponent):

        result = format_str
        result = result.replace(r"%sb", self._get_unit(factor, exponent, 0))
        result = result.replace(r"%lb", self._get_unit(factor, exponent, 1))

        return result

    def get(self, units="autobin", fmt="%0.1f %sb"):
        ''' returns size value as a string with given units and format

            "units" parameter allows to select preferred unit:
                for example "KiB" or "megabytes"
                accepted values are also:
                "autobin" (default) - uses the highest human readable unit (binary)
                "autodec" - uses the highest human readable unit (decimal)

            "fmt" parameter allows to specify the output format:
                %sb - will be replaced with the short byte size unit (e.g. MiB)
                %lb - will be replaced with the long byte size unit (e.g. kilobytes)
                value can be formatted using standard string replacements (e.g. %d, %f)

        '''

        ftr = BINARY_FACTOR
        if units == "autodec":
            ftr = DECIMAL_FACTOR
        if units in ("autobin", "autodec"):
            exp = 0
            value = float(self.bytes)
            while value + 0.01 > ftr:  # + 0.01 to balance the float comparison
                value /= ftr
                exp += 1
        else:
            ftr, exp = self._parse_units(units.strip())
            value = (float(self.factor ** self.exponent) / float(ftr ** exp)) * self.number

        return self._format(fmt, ftr, exp) % value



#!/usr/bin/python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: bsize

short_description: Module for basic manipulation with byte sizes

version_added: "2.5"

description:
    - "Module accepts byte size strings with the units and produces strings in
      form of input accepted by different storage tools"

options:
    size:
        description:
            - String containing number and byte units
        required: true

author:
    - Jan Pokorny (japokorn@redhat.com)
'''

EXAMPLES = '''
# Obtain sizes in format for various tools
- name: Get 10 KiB size
  bsize:
    size: 10 KiB
'''

RETURN = '''
size:
    description: Size in binary format units
    type: str
bytes:
    description: Size in bytes
    type: int
lvm:
    description: Size in binary format. No space after the number,
                 first letter of unit prefix in lowercase only
    type: str
parted:
    description: Size in binary format. No space after the number
    type: str
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.size import Size

def run_module():
    # available arguments/parameters that a user can pass
    module_args = dict(
        size=dict(type='str', required=True),
    )

    # seed the result dict in the object
    result = dict(
        changed=False
    )

    module = AnsibleModule(argument_spec=module_args,
                           supports_check_mode=True)

    size = Size(module.params['size'])

    result['size'] = size.get(fmt="%d %sb")
    result['bytes'] = size.bytes
    result['lvm'] = size.get(fmt="%d%sb").lower()[:-2]
    result['parted'] = size.get(fmt="%d%sb")

    # use whatever logic you need to determine whether or not this module
    # made any modifications to your target
    result['changed'] = False

    # success - return result
    module.exit_json(**result)

def main():
    run_module()

if __name__ == '__main__':
    main()



import pytest
import find_unused_disk
import os


blkid_data_pttype = [('/dev/sdx', '/dev/sdx: PTTYPE=\"dos\"'),
                     ('/dev/sdy', '/dev/sdy: PTTYPE=\"test\"')]

blkid_data = [('/dev/sdx', 'UUID=\"hello-1234-56789\" TYPE=\"crypto_LUKS\"'),
              ('/dev/sdy', 'UUID=\"this-1s-a-t3st-f0r-ansible\" VERSION=\"LVM2 001\" TYPE=\"LVM2_member\" USAGE=\"raid\"'),
              ('/dev/sdz', 'LABEL=\"/data\" UUID=\"a12bcdef-345g-67h8-90i1-234j56789k10\" VERSION=\"1.0\" TYPE=\"ext4\" USAGE=\"filesystem\"')]

holders_data_none = [('/dev/sdx', ''),
                     ('/dev/dm-99', '')]

holders_data = [('/dev/sdx', 'dm-0'),
                ('/dev/dm-99', 'dm-2 dm-3 dm-4')]


@pytest.mark.parametrize('disk, blkid', blkid_data_pttype)
def test_no_signature_true(disk, blkid):
    def run_command(args):
        return [0, blkid, '']
    assert find_unused_disk.no_signature(run_command, disk) is True


@pytest.mark.parametrize('disk, blkid', blkid_data)
def test_no_signature_false(disk, blkid):
    def run_command(args):
        return [0, blkid, '']
    assert find_unused_disk.no_signature(run_command, disk) is False


@pytest.mark.parametrize('disk, holders', holders_data_none)
def test_no_holders_true(disk, holders, monkeypatch):
    def mock_return(args):
        return holders
    monkeypatch.setattr(os, 'listdir', mock_return)
    assert find_unused_disk.no_holders(disk) is True


@pytest.mark.parametrize('disk, holders', holders_data)
def test_no_holders_false(disk, holders, monkeypatch):
    def mock_return(args):
        return holders
    monkeypatch.setattr(os, 'listdir', mock_return)
    assert find_unused_disk.no_holders(disk) is False


def test_can_open_true(monkeypatch):
    def mock_return(args, flag):
        return True
    monkeypatch.setattr(os, 'open', mock_return)
    assert find_unused_disk.can_open('/hello') is True


def test_can_open_false(monkeypatch):
    def mock_return(args, flag):
        raise OSError
    monkeypatch.setattr(os, 'open', mock_return)
    assert find_unused_disk.can_open('/hello') is False



#!/usr/bin/python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: blockdev_info
short_description: Collect info about block devices in the system.
version_added: "2.5"
description:
    - "This module collects information about block devices"
options:
author:
    - David Lehman (dlehman@redhat.com)
'''

EXAMPLES = '''
- name: Get info about block devices
  blockdev_info:
  register: blk_info

'''

RETURN = '''
info:
    description: dict w/ device path keys and device info dict values
    type: dict
'''

import shlex

from ansible.module_utils.basic import AnsibleModule


LSBLK_DEVICE_TYPES = {"part": "partition"}

def get_block_info(run_cmd):
    buf = run_cmd(["lsblk", "-o", "NAME,FSTYPE,LABEL,UUID,TYPE", "-p", "-P", "-a"])[1]
    info = dict()
    for line in buf.splitlines():
        dev = dict()
        for pair in shlex.split(line):
            try:
                key, _eq, value = pair.partition("=")
            except ValueError:
                print(pair)
                raise
            if key:
                dev[key.lower()] = LSBLK_DEVICE_TYPES.get(value, value)
        if dev:
            info[dev['name']] = dev

    return info


def run_module():
    module_args = dict()

    result = dict(
        info=None,
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    try:
        result['info'] = get_block_info(module.run_command)
    except Exception:
        module.fail_json(msg="Failed to collect block device info.")

    module.exit_json(**result)


def main():
    run_module()


if __name__ == '__main__':
    main()



#!/usr/bin/python
"""This module tests methods defined in the lvm_gensym.py module using the pytest framework"""
import pytest


import lvm_gensym


used_lv_names = ['root', 'root_0', 'root_1', 'root_2', 'root_3', 'swap_0', 'swap', 'swap_1']

test_lv_names = [{'fs_type': 'ext', 'mount': '/'},
                 {'fs_type': 'zfs', 'mount': '/home/user'},
                 {'fs_type': 'swap', 'mount': ''}
                ]

used_vg_names = ['linux_host', 'rhel_user0', 'rhel_0_user']

test_vg_names = ['rhel_user', 'rhel_user_0', 'rhel_user_1',
                 'rhel_user_2', 'rhel_user_3', 'linux_user',
                 'fedora_user', 'fedora_user_0', 'fedora_user_1'
                ]

lvm_facts = {'lvs': {'Home': '', 'Swap': '', 'Root': '',
                     'Root_0': '', 'root': '', 'root_0': '',
                     'swap': '', 'swap_0': '', 'swap_1': '',
                    },
             'vgs': {'rhel_user': '', 'rhel_user_0': '', 'rhel_user_1': ''}
            }

def test_unique_base_name():
    """Test whether the returned name is unique using a supplied list of test names"""
    assert lvm_gensym.get_unique_name_from_base('root', used_lv_names) == 'root_4'
    assert lvm_gensym.get_unique_name_from_base('rhel_user', test_vg_names) == 'rhel_user_4'

def test_return_val():
    """Verify that a supplied unique name and a list of used names returns True"""
    for (index, name) in enumerate(test_vg_names):
        assert lvm_gensym.name_is_unique(name[index], used_vg_names)

def test_get_base_vg_name():
    """Check generated base volume group name against the expected base name"""
    assert lvm_gensym.get_vg_name_base('hostname', 'rhel') == 'rhel_hostname'

@pytest.mark.parametrize("os_name", ["foo", "bar", "baz"])
def test_vg_eval(monkeypatch, os_name):
    """Check generated unique volume group name against the expected name"""
    def get_os_name():
        return os_name

    vg_names = [os_name + "_user", os_name + "_user_0", os_name + "_user_1"]
    _lvm_facts = dict(vgs=dict.fromkeys(vg_names), lvs=dict())
    monkeypatch.setattr(lvm_gensym, "get_os_name", get_os_name)
    assert lvm_gensym.get_vg_name('user', _lvm_facts) == os_name + '_user_2'
    assert lvm_gensym.get_vg_name('', _lvm_facts) == os_name

def test_lv_eval():
    """Test the generated unique logical volume name against the expected name"""
    expected = ['root_1', 'home_user', 'swap_2']

    for (ctr, name_inputs) in enumerate(test_lv_names):
        assert lvm_gensym.get_lv_name(name_inputs['fs_type'], name_inputs['mount'], lvm_facts) == expected[ctr]

def test_get_base_lv_name():
    """Test the generated base logical volume name against the expected name"""
    expected = ['root', 'home_user', 'swap']

    for (ctr, names_input) in enumerate(test_lv_names):
        assert lvm_gensym.get_lv_name_base(names_input['fs_type'], names_input['mount']) == expected[ctr]



#!/usr/bin/python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: blivet

short_description: Module for management of linux block device stacks

version_added: "2.5"

description:
    - "Module configures storage pools and volumes to match the state specified
       in input parameters. It does not do any management of /etc/fstab entries."

options:
    pools:
        description:
            - list of dicts describing pools
    volumes:
        description:
            - list of dicts describing volumes
    use_partitions:
        description:
            - boolean indicating whether to create partitions on disks for pool backing devices
    disklabel_type:
        description:
            - disklabel type string (eg: 'gpt') to use, overriding the built-in logic in blivet

author:
    - David Lehman (dlehman@redhat.com)
'''

EXAMPLES = '''

- name: Manage devices
  blivet:
    pools:
        - name: "{{ app_pool }}"
          disks: ["sdd", "sde"]
          volumes:
            - name: shared
              size: "10 GiB"
              mount_point: /opt/{{ app_pool }}/shared
            - name: web
              size: 8g
              mount_point: /opt/{{ app_pool }}/web
    volumes:
        - name: whole_disk1
          disks: ['sdc']
          mount_point: /whole_disk1
          fs_type: ext4
          mount_options: journal_checksum,async,noexec
'''

RETURN = '''
actions:
    description: list of dicts describing actions taken
    type: list of dict
leaves:
    description: list of paths to leaf devices
    type: list of str
mounts:
    description: list of dicts describing mounts to set up
    type: list of dict
pools:
    description: list of dicts describing the pools w/ device path for each volume
    type: list of dict
volumes:
    description: list of dicts describing the volumes w/ device path for each
    type: list of dict
'''

import logging
import traceback

BLIVET_PACKAGE = None

try:
    from blivet3 import Blivet
    from blivet3.callbacks import callbacks
    from blivet3.flags import flags as blivet_flags
    from blivet3.formats import get_format
    from blivet3.partitioning import do_partitioning
    from blivet3.size import Size
    from blivet3.util import set_up_logging
    BLIVET_PACKAGE = 'blivet3'
except ImportError:
    try:
        from blivet import Blivet
        from blivet.callbacks import callbacks
        from blivet.flags import flags as blivet_flags
        from blivet.formats import get_format
        from blivet.partitioning import do_partitioning
        from blivet.size import Size
        from blivet.util import set_up_logging
        BLIVET_PACKAGE = 'blivet'
    except ImportError:
        LIB_IMP_ERR = traceback.format_exc()

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

if BLIVET_PACKAGE:
    blivet_flags.debug = True
    set_up_logging()
    log = logging.getLogger(BLIVET_PACKAGE + ".ansible")

use_partitions = None  # create partitions on pool backing device disks?
disklabel_type = None  # user-specified disklabel type


class BlivetAnsibleError(Exception):
    pass


class BlivetVolume:
    def __init__(self, blivet_obj, volume, bpool=None):
        self._blivet = blivet_obj
        self._volume = volume
        self._blivet_pool = bpool
        self._device = None

    @property
    def ultimately_present(self):
        """ Should this volume be present when we are finished? """
        return (self._volume['state'] == 'present' and
                (self._blivet_pool is None or self._blivet_pool.ultimately_present))

    def _type_check(self):  # pylint: disable=no-self-use
        """ Is self._device of the correct type? """
        return True

    def _get_device_id(self):
        """ Return an identifier by which to try looking the volume up. """
        return self._volume['name']

    def _look_up_device(self):
        """ Try to look up this volume in blivet's device tree. """
        device = self._blivet.devicetree.resolve_device(self._get_device_id())
        if device is None:
            return

        self._device = device

        # check that the type is correct, raising an exception if there is a name conflict
        if not self._type_check():
            self._device = None
            return  # TODO: see if we can create this device w/ the specified name

    def _get_format(self):
        """ Return a blivet.formats.DeviceFormat instance for this volume. """
        fmt = get_format(self._volume['fs_type'],
                         mountpoint=self._volume.get('mount_point'),
                         label=self._volume['fs_label'],
                         options=self._volume['fs_create_options'])
        if not fmt.supported or not fmt.formattable:
            # FAIL: fs type tools are not available
            raise BlivetAnsibleError("required tools for file system '%s' are missing" % self._volume['fs_type'])

        return fmt

    def _create(self):
        """ Schedule actions as needed to ensure the volume exists. """
        pass

    def _destroy(self):
        """ Schedule actions as needed to ensure the volume does not exist. """
        if self._device is None:
            return

        # save device identifiers for use by the role
        self._volume['_device'] = self._device.path
        self._volume['_mount_id'] = self._device.fstab_spec

        # schedule removal of this device and any descendant devices
        self._blivet.devicetree.recursive_remove(self._device)

    def _resize(self):
        """ Schedule actions as needed to ensure the device has the desired size. """
        try:
            size = Size(self._volume['size'])
        except Exception:
            # FAIL: invalid size specification
            raise BlivetAnsibleError("invalid size specification for volume '%s': '%s'" % (self._volume['name'], self._volume['size']))

        if size and self._device.resizable and self._device.size != size:
            if self._device.format.resizable:
                self._device.format.update_size_info()

            if not self._device.min_size <= size <= self._device.max_size:
                # FAIL: resize to specified size not possible
                raise BlivetAnsibleError("volume '%s' cannot be resized to '%s'" % (self._volume['name'], size))

            try:
                self._blivet.resize_device(self._device, size)
            except ValueError as e:
                # FAIL: resize not possible
                raise BlivetAnsibleError("volume '%s' cannot be resized from %s to %s: %s" % (self._device.name,
                                                                                              self._device.size,
                                                                                              size, str(e)))

    def _reformat(self):
        """ Schedule actions as needed to ensure the volume is formatted as specified. """
        fmt = self._get_format()
        if self._device.format.type == fmt.type:
            return

        if self._device.format.status:
            self._device.format.teardown()
        self._blivet.format_device(self._device, fmt)

    def manage(self):
        """ Schedule actions to configure this volume according to the yaml input. """
        # look up the device
        self._look_up_device()

        # schedule destroy if appropriate
        if not self.ultimately_present:
            self._destroy()
            return

        # schedule create if appropriate
        self._create()

        # at this point we should have a blivet.devices.StorageDevice instance
        if self._device is None:
            raise BlivetAnsibleError("failed to look up or create device '%s'" % self._volume['name'])

        # schedule reformat if appropriate
        if self._device.exists:
            self._reformat()

        # schedule resize if appropriate
        if self._device.exists and self._volume['size']:
            self._resize()

        # save device identifiers for use by the role
        self._volume['_device'] = self._device.path
        self._volume['_mount_id'] = self._device.fstab_spec


class BlivetDiskVolume(BlivetVolume):
    def _get_device_id(self):
        return self._volume['disks'][0]

    def _type_check(self):
        return self._device.is_disk


class BlivetPartitionVolume(BlivetVolume):
    def _type_check(self):
        return self._device.type == 'partition'

    def _get_device_id(self):
        return self._blivet_pool._disks[0].name + '1'

    def _create(self):
        if self._device:
            return

        if self._blivet_pool:
            parent = self._blivet_pool._device
        else:
            parent = self._blivet.devicetree.resolve_device(self._volume['pool'])

        if parent is None:
            # FAIL: failed to find pool
            raise BlivetAnsibleError("failed to find pool '%s' for volume '%s'" % (self._blivet_pool['name'], self._volume['name']))

        size = Size("256 MiB")
        try:
            device = self._blivet.new_partition(parents=[parent], size=size, grow=True, fmt=self._get_format())
        except Exception:
            # FAIL: failed to instantiate volume device
            raise BlivetAnsibleError("failed set up volume '%s'" % self._volume['name'])

        self._blivet.create_device(device)
        try:
            do_partitioning(self._blivet)
        except Exception:
            # FAIL: partition allocation failed: not enough space?
            raise BlivetAnsibleError("partition allocation failed for volume '%s'" % self._volume['name'])

        self._device = device


class BlivetLVMVolume(BlivetVolume):
    def _get_device_id(self):
        return "%s-%s" % (self._blivet_pool._device.name, self._volume['name'])

    def _create(self):
        if self._device:
            return

        parent = self._blivet_pool._device
        if parent is None:
            # FAIL: failed to find pool
            raise BlivetAnsibleError("failed to find pool '%s' for volume '%s'" % (self._blivet_pool['name'], self._volume['name']))

        try:
            size = Size(self._volume['size'])
        except Exception:
            # FAIL: invalid size specification
            raise BlivetAnsibleError("invalid size '%s' specified for volume '%s'" % (self._volume['size'], self._volume['name']))

        fmt = self._get_format()
        if size > parent.free_space:
            # FAIL: volume size greater than pool free space
            raise BlivetAnsibleError("specified size for volume '%s' exceeds available space in pool '%s' (%s)" % (size,
                                                                                                                   parent.name,
                                                                                                                   parent.free_space))

        try:
            device = self._blivet.new_lv(name=self._volume['name'],
                                         parents=[parent], size=size, fmt=fmt)
        except Exception:
            # FAIL: failed to create volume
            raise BlivetAnsibleError("failed to set up volume '%s'" % self._volume['name'])

        self._blivet.create_device(device)
        self._device = device


_BLIVET_VOLUME_TYPES = {
    "disk": BlivetDiskVolume,
    "lvm": BlivetLVMVolume,
    "partition": BlivetPartitionVolume
}


def _get_blivet_volume(blivet_obj, volume, bpool=None):
    """ Return a BlivetVolume instance appropriate for the volume dict. """
    volume_type = volume.get('type', bpool._pool['type'] if bpool else None)
    if volume_type not in _BLIVET_VOLUME_TYPES:
        raise BlivetAnsibleError("Volume '%s' has unknown type '%s'" % (volume['name'], volume_type))

    return _BLIVET_VOLUME_TYPES[volume_type](blivet_obj, volume, bpool=bpool)


class BlivetPool:
    def __init__(self, blivet_obj, pool):
        self._blivet = blivet_obj
        self._pool = pool
        self._device = None
        self._disks = list()
        self._blivet_volumes = list()

    @property
    def ultimately_present(self):
        """ Should this pool be present when we are finished? """
        return self._pool['state'] == 'present'

    def _create(self):
        """ Schedule actions as needed to ensure the pool exists. """
        pass

    def _destroy(self):
        """ Schedule actions as needed to ensure the pool does not exist. """
        if self._device is None:
            return

        ancestors = self._device.ancestors  # ascending distance ordering
        log.debug("%s", [a.name for a in ancestors])
        self._blivet.devicetree.recursive_remove(self._device)
        ancestors.remove(self._device)
        leaves = [a for a in ancestors if a.isleaf]
        while leaves:
            for ancestor in leaves:
                log.info("scheduling destruction of %s", ancestor.name)
                if ancestor.is_disk:
                    self._blivet.devicetree.recursive_remove(ancestor)
                else:
                    self._blivet.destroy_device(ancestor)

                ancestors.remove(ancestor)

            leaves = [a for a in ancestors if a.isleaf]

    def _type_check(self):  # pylint: disable=no-self-use
        return True

    def _look_up_disks(self):
        """ Look up the pool's disks in blivet's device tree. """
        if not self._pool['disks']:
            # FAIL: no disks specified for pool
            raise BlivetAnsibleError("no disks specified for pool '%s'" % self._pool['name'])  # sure about this one?
        elif not isinstance(self._pool['disks'], list):
            raise BlivetAnsibleError("pool disks must be specified as a list")

        disks = list()
        for spec in self._pool['disks']:
            device = self._blivet.devicetree.resolve_device(spec)
            if device is not None:  # XXX fail if any disk isn't resolved?
                disks.append(device)

        if self._pool['disks'] and not disks:
            # FAIL: failed to find any disks
            raise BlivetAnsibleError("unable to resolve any disks specified for pool '%s' (%s)" % (self._pool['name'], self._pool['disks']))

        self._disks = disks

    def _look_up_device(self):
        """ Look up the pool in blivet's device tree. """
        device = self._blivet.devicetree.resolve_device(self._pool['name'])
        if device is None:
            return

        self._device = device

        # check that the type is correct, raising an exception if there is a name conflict
        if not self._type_check():
            self._device = None
            return  # TODO: see if we can create this device w/ the specified name

    def _get_format(self):
        raise NotImplementedError()

    def _create_members(self):
        """ Schedule actions as needed to ensure pool member devices exist. """
        members = list()
        for disk in self._disks:
            if not disk.isleaf:
                self._blivet.devicetree.recursive_remove(disk)

            if use_partitions:
                label = get_format("disklabel", device=disk.path)
                self._blivet.format_device(disk, label)
                member = self._blivet.new_partition(parents=[disk], size=Size("256MiB"), grow=True)
                self._blivet.create_device(member)
            else:
                member = disk

            self._blivet.format_device(member, self._get_format())
            members.append(member)

        if use_partitions:
            try:
                do_partitioning(self._blivet)
            except Exception:
                # FAIL: problem allocating partitions for pool backing devices
                raise BlivetAnsibleError("failed to allocation partitions for pool '%s'" % self._pool['name'])

        return members

    def _get_volumes(self):
        """ Set up BlivetVolume instances for this pool's volumes. """
        for volume in self._pool['volumes']:
            bvolume = _get_blivet_volume(self._blivet, volume, self)
            self._blivet_volumes.append(bvolume)

    def _manage_volumes(self):
        """ Schedule actions as needed to configure this pool's volumes. """
        self._get_volumes()
        for bvolume in self._blivet_volumes:
            bvolume.manage()

    def manage(self):
        """ Schedule actions to configure this pool according to the yaml input. """
        # look up the device
        self._look_up_disks()
        self._look_up_device()

        # schedule destroy if appropriate, including member type change
        if not self.ultimately_present:  # TODO: member type changes
            self._manage_volumes()
            self._destroy()
            return

        # schedule create if appropriate
        self._create()
        self._manage_volumes()


class BlivetPartitionPool(BlivetPool):
    def _type_check(self):
        return self._device.partitionable

    def _look_up_device(self):
        self._device = self._disks[0]

    def _create(self):
        if self._device.format.type != "disklabel" or \
           self._device.format.label_type != disklabel_type:
            self._blivet.devicetree.recursive_remove(self._device, remove_device=False)

            label = get_format("disklabel", device=self._device.path, label_type=disklabel_type)
            self._blivet.format_device(self._device, label)


class BlivetLVMPool(BlivetPool):
    def _type_check(self):
        return self._device.type == "lvmvg"

    def _get_format(self):
        fmt = get_format("lvmpv")
        if not fmt.supported or not fmt.formattable:
            # FAIL: lvm tools are not available
            raise BlivetAnsibleError("required tools for managing LVM are missing")

        return fmt

    def _create(self):
        if self._device:
            return

        members = self._create_members()
        try:
            pool_device = self._blivet.new_vg(name=self._pool['name'], parents=members)
        except Exception:
            # FAIL: failed to instantiate pool device
            raise BlivetAnsibleError("failed to set up pool '%s'" % self._pool['name'])

        self._blivet.create_device(pool_device)
        self._device = pool_device


_BLIVET_POOL_TYPES = {
    "disk": BlivetPartitionPool,
    "lvm": BlivetLVMPool
}


def _get_blivet_pool(blivet_obj, pool):
    """ Return an appropriate BlivetPool instance for the pool dict. """
    if pool['type'] not in _BLIVET_POOL_TYPES:
        raise BlivetAnsibleError("Pool '%s' has unknown type '%s'" % (pool['name'], pool['type']))

    return _BLIVET_POOL_TYPES[pool['type']](blivet_obj, pool)


def manage_volume(b, volume):
    """ Schedule actions as needed to manage a single standalone volume. """
    bvolume = _get_blivet_volume(b, volume)
    bvolume.manage()
    volume['_device'] = bvolume._volume.get('_device', '')
    volume['_mount_id'] = bvolume._volume.get('_mount_id', '')


def manage_pool(b, pool):
    """ Schedule actions as needed to manage a single pool and its volumes. """
    bpool = _get_blivet_pool(b, pool)
    bpool.manage()
    for (volume, bvolume) in zip(pool['volumes'], bpool._blivet_volumes):
        volume['_device'] = bvolume._volume.get('_device', '')
        volume['_mount_id'] = bvolume._volume.get('_mount_id', '')


class FSTab:
    def __init__(self, blivet_obj):
        self._blivet = blivet_obj
        self._entries = list()
        self.parse()

    def lookup(self, key, value):
        return next((e for e in self._entries if e.get(key) == value), None)

    def reset(self):
        self._entries = list()

    def parse(self):
        if self._entries:
            self.reset()

        for line in open('/etc/fstab').readlines():
            if line.lstrip().startswith("#"):
                continue

            fields = line.split()
            if len(fields) < 6:
                continue

            device = self._blivet.devicetree.resolve_device(fields[0])
            self._entries.append(dict(device_id=fields[0],
                                      device_path=getattr(device, 'path', None),
                                      fs_type=fields[2],
                                      mount_point=fields[1],
                                      mount_options=fields[3]))


def get_mount_info(pools, volumes, actions, fstab):
    """ Return a list of argument dicts to pass to the mount module to manage mounts.

        The overall approach is to remove existing mounts associated with file systems
        we are removing and those with changed mount points, re-adding them with the
        new mount point later.

        Removed mounts go directly into the mount_info list, which is the return value,
        while added/active mounts to a list that gets appended to the mount_info list
        at the end to ensure that removals happen first.
    """
    mount_info = list()
    mount_vols = list()

    # account for mounts removed by removing or reformatting volumes
    if actions:
        for action in actions:
            if action.is_destroy and action.is_format and action.format.type is not None:
                mount = fstab.lookup('device_path', action.device.path)
                if mount is not None:
                    mount_info.append({"path": mount['mount_point'], 'state': 'absent'})

    def handle_new_mount(volume, fstab):
        replace = None
        mounted = False

        mount = fstab.lookup('device_path', volume['_device'])
        if volume['mount_point']:
            mounted = True

        # handle removal of existing mounts of this volume
        if mount and mount['mount_point'] != volume['mount_point']:
            replace = mount['mount_point']

        return mounted, replace

    # account for mounts that we set up or are replacing in pools
    for pool in pools:
        for volume in pool['volumes']:
            if pool['state'] == 'present' and volume['state'] == 'present':
                mounted, replace = handle_new_mount(volume, fstab)
                if replace:
                    mount_info.append({"path": replace, 'state': 'absent'})
                if mounted:
                    mount_vols.append(volume)

    # account for mounts that we set up or are replacing in standalone volumes
    for volume in volumes:
        if volume['state'] == 'present':
            mounted, replace = handle_new_mount(volume, fstab)
            if replace:
                mount_info.append({"path": replace, 'state': 'absent'})
            if mounted:
                mount_vols.append(volume)

    for volume in mount_vols:
        mount_info.append({'src': volume['_mount_id'],
                           'path': volume['mount_point'],
                           'fstype': volume['fs_type'],
                           'opts': volume['mount_options'],
                           'dump': volume['mount_check'],
                           'passno': volume['mount_passno'],
                           'state': 'mounted'})

    return mount_info


def run_module():
    # available arguments/parameters that a user can pass
    module_args = dict(
        pools=dict(type='list'),
        volumes=dict(type='list'),
        packages_only=dict(type='bool', required=False, default=False),
        disklabel_type=dict(type='str', required=False, default=None),
        use_partitions=dict(type='bool', required=False, default=True))

    # seed the result dict in the object
    result = dict(
        changed=False,
        actions=list(),
        leaves=list(),
        mounts=list(),
        pools=list(),
        volumes=list(),
        packages=list(),
    )

    module = AnsibleModule(argument_spec=module_args,
                           supports_check_mode=True)
    if not BLIVET_PACKAGE:
        module.fail_json(msg=missing_required_lib("blivet"),
                         exception=LIB_IMP_ERR)

    if not module.params['pools'] and not module.params['volumes']:
        module.exit_json(**result)

    global disklabel_type
    disklabel_type = module.params['disklabel_type']

    global use_partitions
    use_partitions = module.params['use_partitions']

    b = Blivet()
    b.reset()
    fstab = FSTab(b)
    actions = list()

    def record_action(action):
        if action.is_format and action.format.type is None:
            return

        actions.append(action)

    def action_dict(action):
        return dict(action=action.type_desc_str,
                    fs_type=action.format.type if action.is_format else None,
                    device=action.device.path)

    for pool in module.params['pools']:
        try:
            manage_pool(b, pool)
        except BlivetAnsibleError as e:
            module.fail_json(msg=str(e), **result)

    for volume in module.params['volumes']:
        try:
            manage_volume(b, volume)
        except BlivetAnsibleError as e:
            module.fail_json(msg=str(e), **result)

    scheduled = b.devicetree.actions.find()
    result['packages'] = b.packages[:]

    if module.params['packages_only']:
        module.exit_json(**result)

    for action in scheduled:
        if action.is_destroy and action.is_format and action.format.exists:
            action.format.teardown()

    if scheduled:
        ## execute the scheduled actions, committing changes to disk
        callbacks.action_executed.add(record_action)
        try:
            b.devicetree.actions.process(devices=b.devicetree.devices, dry_run=module.check_mode)
        except Exception:
            module.fail_json(msg="Failed to commit changes to disk", **result)
        finally:
            result['changed'] = True
            result['actions'] = [action_dict(a) for a in actions]

    result['mounts'] = get_mount_info(module.params['pools'], module.params['volumes'], actions, fstab)
    result['leaves'] = [d.path for d in b.devicetree.leaves]
    result['pools'] = module.params['pools']
    result['volumes'] = module.params['volumes']

    # success - return result
    module.exit_json(**result)

def main():
    run_module()

if __name__ == '__main__':
    main()
